@@ -1,4 +1,4 @@
-## Copyright (C) 2003-2004 Rubens Ramos <rubensr@users.sourceforge.net>
+## Copyright (C) 2003-2006 Rubens Ramos <rubensr@users.sourceforge.net>
 
 ## Based on code by:
 ## Copyright (C) 2003  Razvan Cojocaru <razvanco@gmx.net>
@@ -18,7 +18,7 @@
 ## write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 ## Boston, MA 02111-1307, USA
 
-## $Id: chm.py,v 1.7 2004/11/27 04:51:28 rubensr Exp $
+## $Id: chm.py,v 1.10 2006/04/08 07:40:47 rubensr Exp $
 
 '''
    chm - A high-level front end for the chmlib python module.
@@ -45,7 +45,7 @@ charset_table = {
     162 : 'iso8859_9',  # TURKISH_CHARSET
     222 : 'iso8859_11', # THAI_CHARSET - hmm not in python 2.2...
     186 : 'iso8859_13', # BALTIC_CHARSET
-    204 : 'koi8',       # RUSSIAN_CHARSET
+    204 : 'cp1251',     # RUSSIAN_CHARSET
     255 : 'cp437',      # OEM_CHARSET
     128 : 'cp932',      # SHIFTJIS_CHARSET
     134 : 'cp936',      # GB2312_CHARSET
@@ -79,10 +79,10 @@ locale_table = {
     0x4001 : ('iso8859_6', "Arabic_Qatar", "Arabic"),
     0x042b : (None,        "Armenian","Armenian"),
     0x042c : ('iso8859_9', "Azeri_Latin", "Turkish"),
-    0x082c : ('koi8',      "Azeri_Cyrillic", "Cyrillic"),
+    0x082c : ('cp1251',    "Azeri_Cyrillic", "Cyrillic"),
     0x042d : ('iso8859_1', "Basque", "Western Europe & US"),
-    0x0423 : ('koi8',      "Belarusian", "Cyrillic"),
-    0x0402 : ('koi8',      "Bulgarian", "Cyrillic"),
+    0x0423 : ('cp1251',    "Belarusian", "Cyrillic"),
+    0x0402 : ('cp1251',    "Bulgarian", "Cyrillic"),
     0x0403 : ('iso8859_1', "Catalan", "Western Europe & US"),
     0x0404 : ('cp950',     "Chinese_Taiwan", "Traditional Chinese"),
     0x0804 : ('cp936',     "Chinese_PRC", "Simplified Chinese"),
@@ -132,12 +132,12 @@ locale_table = {
     0x0410 : ('iso8859_1', "Italian_Standard", "Western Europe & US"),
     0x0810 : ('iso8859_1', "Italian_Swiss", "Western Europe & US"),
     0x0411 : ('cp932',     "Japanese", "Japanese"),
-    0x043f : ('koi8',      "Kazakh", "Cyrillic"),
+    0x043f : ('cp1251',    "Kazakh", "Cyrillic"),
     0x0457 : (None,        "Konkani", "Indic"),
     0x0412 : ('cp949',     "Korean", "Korean"),
     0x0426 : ('iso8859_13',"Latvian", "Baltic",),
     0x0427 : ('iso8859_13',"Lithuanian", "Baltic",),
-    0x042f : ('koi8',      "Macedonian", "Cyrillic"),
+    0x042f : ('cp1251',    "Macedonian", "Cyrillic"),
     0x043e : ('iso8859_1', "Malay_Malaysia", "Western Europe & US"),
     0x083e : ('iso8859_1', "Malay_Brunei_Darussalam", "Western Europe & US"),
     0x044e : (None,        "Marathi", "Indic"),
@@ -147,10 +147,10 @@ locale_table = {
     0x0416 : ('iso8859_1', "Portuguese_Brazilian", "Western Europe & US"),
     0x0816 : ('iso8859_1', "Portuguese_Standard", "Western Europe & US"),
     0x0418 : ('iso8859_2', "Romanian", "Central Europe"),
-    0x0419 : ('koi8',      "Russian", "Cyrillic"),
+    0x0419 : ('cp1251',    "Russian", "Cyrillic"),
     0x044f : (None,        "Sanskrit", "Indic"),
     0x081a : ('iso8859_2', "Serbian_Latin", "Central Europe"),
-    0x0c1a : ('koi8',      "Serbian_Cyrillic", "Cyrillic"),
+    0x0c1a : ('cp1251',    "Serbian_Cyrillic", "Cyrillic"),
     0x041b : ('iso8859_2', "Slovak", "Central Europe"),
     0x0424 : ('iso8859_2', "Slovenian", "Central Europe"),
     0x040a : ('iso8859_1', "Spanish_Trad_Sort", "Western Europe & US"),
@@ -177,13 +177,13 @@ locale_table = {
     0x041d : ('iso8859_1', "Swedish", "Western Europe & US"),
     0x081d : ('iso8859_1', "Swedish_Finland", "Western Europe & US"),
     0x0449 : (None,        "Tamil", "Indic"),
-    0x0444 : ('koi8',      "Tatar", "Cyrillic"),
+    0x0444 : ('cp1251',    "Tatar", "Cyrillic"),
     0x041e : ('iso8859_11',"Thai", "Thai"),
     0x041f : ('iso8859_9', "Turkish", "Turkish"),
-    0x0422 : ('koi8',      "Ukrainian", "Cyrillic"),
+    0x0422 : ('cp1251',    "Ukrainian", "Cyrillic"),
     0x0420 : ('iso8859_6', "Urdu", "Arabic"),
     0x0443 : ('iso8859_9', "Uzbek_Latin", "Turkish"),
-    0x0843 : ('koi8',      "Uzbek_Cyrillic", "Cyrillic"),
+    0x0843 : ('cp1251',    "Uzbek_Cyrillic", "Cyrillic"),
     0x042a : (None,        "Vietnamese", "Vietnamese")
 }
 
@@ -312,29 +312,9 @@ class CHMFile:
                 cursor = buff[index] + (buff[index+1] * 256)
                 index += 2
             index += cursor
-        if ((not self.topics) and (not self.index)):
-            result, ui = chmlib.chm_resolve_object(self.file, '/#STRINGS')
-            if (result != chmlib.CHM_RESOLVE_SUCCESS):
-                print 'GetArchiveInfo: Could not find #STRINGS'
-                return 1
-            
-            size, text = chmlib.chm_retrieve_object(self.file, ui, 1l,
-                                                    ui.length)
-            if (size == 0):
-                print 'GetArchiveInfo: STRINGS file size = 0'
-                return 1
-
-            index = 0
-            while (index < size):
-                next = string.find(text, '\x00', index)
-                chunk = text[index:next]
-                ext = string.lower(chunk[-4:])
-                if (ext == '.hhc'):
-                    self.topics = '/' + chunk
-                elif (ext == '.hhk'):
-                   self.index = '/' + chunk
-                index = next + 1
 
+        self.GetWindowsInfo()
+        
         if not self.lcid:
             self.lcid = extra.get_lcid (self.file)
 
@@ -400,8 +380,12 @@ class CHMFile:
         if self.file and ui:
             if length == -1:
                 len = ui.length
+            else:
+                len = length
             if start == -1:
                 st = 0l
+            else:
+                st = long(start)
             return chmlib.chm_retrieve_object(self.file, ui, st, len)
         else:
             return (0, '')
@@ -447,3 +431,70 @@ class CHMFile:
         else:
             return None
 
+    def GetDWORD(self, buff, idx=0):
+        '''Internal method.
+        Reads a double word (4 bytes) from a buffer.
+        '''
+        result = buff[idx] + (buff[idx+1]<<8) + (buff[idx+2]<<16) + \
+                 (buff[idx+3]<<24)
+
+        if result == 0xFFFFFFFF:
+            result = 0
+            
+        return result
+
+    def GetString(self, text, idx):
+        '''Internal method.
+        Retrieves a string from the #STRINGS buffer.
+        '''
+        next = string.find(text, '\x00', idx)
+        chunk = text[idx:next]
+        return chunk
+    
+    def GetWindowsInfo(self):
+        '''Gets information from the #WINDOWS file.
+        Checks the #WINDOWS file to see if it has any info that was
+        not found in #SYSTEM (topics, index or default page.
+        '''
+        result, ui = chmlib.chm_resolve_object(self.file, '/#WINDOWS')
+        if (result != chmlib.CHM_RESOLVE_SUCCESS):
+            return -1
+
+        size, text = chmlib.chm_retrieve_object(self.file, ui, 0l, 8)
+        if (size < 8):
+            return -2
+
+        buff = array.array('B', text)
+        num_entries = self.GetDWORD(buff, 0)
+        entry_size = self.GetDWORD(buff, 4)
+
+        if num_entries < 1:
+            return -3
+        
+        size, text = chmlib.chm_retrieve_object(self.file, ui, 8l, entry_size)
+        if (size < entry_size):
+            return -4
+
+        buff = array.array('B', text)
+        toc_index = self.GetDWORD(buff, 0x60)
+        idx_index = self.GetDWORD(buff, 0x64)
+        dft_index = self.GetDWORD(buff, 0x68)
+        
+        result, ui = chmlib.chm_resolve_object(self.file, '/#STRINGS')
+        if (result != chmlib.CHM_RESOLVE_SUCCESS):
+            return -5
+        
+        size, text = chmlib.chm_retrieve_object(self.file, ui, 0l, ui.length)
+        if (size == 0):
+            return -6
+
+        if ((not self.topics) and (toc_index != 0)):
+            self.topics = self.GetString(text, toc_index)
+            
+        if ((not self.index) and (idx_index != 0)):
+            self.index = self.GetString(text, idx_index)
+
+        if (dft_index != 0):
+            self.home = self.GetString(text, dft_index)
+            if not self.home.startswith("/"):
+                self.home = "/" + self.home
